@@ -12,7 +12,7 @@ using namespace kmx;
 /* Globals */
 
 KMX_BOOL km::core::kmx::g_debug_ToConsole = FALSE;
-KMX_BOOL km::core::kmx::g_debug_KeymanLog = TRUE;
+KMX_BOOL km::core::kmx::g_debug_KeymanLog = FALSE; // workaround for #12661
 KMX_BOOL km::core::kmx::g_silent = FALSE;
 
 /*
