@@ -122,6 +122,50 @@ bool LdmlTestSource::get_expected_beep() const {
   return false;
 }
 
+int LdmlTestSource::load_kmx_plus(const km::core::path &compiled) {
+
+  // check and load the KMX (yes, once again)
+  if(!km::core::ldml_processor::is_kmxplus_file(compiled, rawdata)) {
+    std::cerr << "Reading KMX for test purposes failed: " << compiled << std::endl;
+    return __LINE__;
+  }
+
+  auto comp_keyboard = (const km::core::kmx::COMP_KEYBOARD*)rawdata.data();
+  // initialize the kmxplus object with our copy
+  kmxplus.reset(new km::core::kmx::kmx_plus(comp_keyboard, rawdata.size()));
+
+  if (!kmxplus->is_valid()) {
+    std::cerr << "kmx_plus invalid" << std::endl;
+    return __LINE__;
+  }
+
+  if (!kmxplus->key2Helper.valid()) {
+    std::cerr << "kmx_plus invalid" << std::endl;
+    return __LINE__;
+  }
+
+  return 0; // success
+}
+
+bool LdmlTestSource::get_vkey_table(std::set<key_event> &fillin) const {
+  if (!kmxplus || !kmxplus->is_valid()) {
+    return false; // fail
+  }
+
+  // just dump the kmap table
+  for (KMX_DWORD kmapIdx = 0; kmapIdx < kmxplus->key2->kmapCount; kmapIdx++) {
+    const km::core::kmx::COMP_KMXPLUS_KEYS_KMAP *kmap = kmxplus->key2Helper.getKmap(kmapIdx);
+    if (kmap == nullptr) {
+      return false;
+    }
+    if (kmap->vkey > 0xFF) {
+      continue; // synthetic key- skip
+    }
+    fillin.insert(key_event(kmap->vkey, kmap->mod));
+  }
+  return true;
+}
+
 // String trim functions from https://stackoverflow.com/a/217605/1836776
 // trim from start (in place)
 static inline void
@@ -259,12 +303,13 @@ LdmlEmbeddedTestSource::is_token(const std::string token, std::string &line) {
 }
 
 int
-LdmlEmbeddedTestSource::load_source( const km::core::path &path ) {
+LdmlEmbeddedTestSource::load_source( const km::core::path &path, const km::core::path &compiled ) {
   const std::string s_keys = "@@keys: ";
   const std::string s_expected = "@@expected: ";
   const std::string s_context = "@@context: ";
   const std::string s_capsLock = "@@capsLock: ";
   const std::string s_expecterror = "@@expect-error: ";
+  const std::string s_keylist = "@@keylist: ";
 
   // Parse out the header statements in file.kmn that tell us (a) environment, (b) key sequence, (c) start context, (d) expected
   // result
@@ -296,6 +341,10 @@ LdmlEmbeddedTestSource::load_source( const km::core::path &path ) {
       context = parse_source_string(line);
     } else if (is_token(s_capsLock, line)) {
       set_caps_lock_on(parse_source_string(line).compare(u"1") == 0);
+    } else if (is_token(s_keylist, line)) {
+      set_keylist(line);
+    } else if (line[0] == '@') {
+      std::cerr << path << " warning, unknown @-command " << line << std::endl;
     }
   }
 
@@ -311,7 +360,13 @@ LdmlEmbeddedTestSource::load_source( const km::core::path &path ) {
     return __LINE__;
   }
 
-  return 0;
+  // and load KMX+ for keylist
+  if (!expected_error) {
+    // don't attempt to load KMX+ on expected error
+    return load_kmx_plus(compiled);
+  } else {
+    return 0;
+  }
 }
 
 km_core_status
@@ -352,7 +407,7 @@ LdmlTestSource::char_to_event(char ch) {
 }
 
 uint16_t
-LdmlTestSource::get_modifier(std::string const m) {
+LdmlTestSource::get_modifier(std::string const &m) {
   for (int i = 0; km::core::kmx::s_modifier_names[i].name; i++) {
     if (m == km::core::kmx::s_modifier_names[i].name) {
       return km::core::kmx::s_modifier_names[i].modifier;
@@ -361,6 +416,12 @@ LdmlTestSource::get_modifier(std::string const m) {
   return 0;
 }
 
+std::string key_event::dump() const {
+  std::stringstream f;
+  f  << "Key: {" << km::core::kmx::Debug_VirtualKey(vk) << ", " << km::core::kmx::Debug_ModifierName(modifier_state) << "}";
+  return f.str();
+}
+
 key_event
 LdmlEmbeddedTestSource::vkey_to_event(std::string const &vk_event) {
   // vkey format is MODIFIER MODIFIER K_NAME
@@ -396,34 +457,37 @@ LdmlEmbeddedTestSource::vkey_to_event(std::string const &vk_event) {
 
 void
 LdmlEmbeddedTestSource::next_action(ldml_action &fillin) {
-  if (is_done || keys.empty()) {
-    // We were already done. return done.
-    fillin.type = LDML_ACTION_DONE;
-    return;
+  if (keys.empty()) {
+    // #3 we are almost done, let's run the key check
+    if (check_keylist) {
+      fillin.type = LDML_ACTION_CHECK_KEYLIST;
+      fillin.string = expected_keylist; // could be empty
+      check_keylist = false;
+    } else {
+      fillin.type = LDML_ACTION_DONE;
+    }
   } else if(keys[0].empty()) {
+    // #2. Then, when we finish a key set, we check the 'expected' at the end of it.
     // Got to the end of a key set. time to check
     fillin.type = LDML_ACTION_CHECK_EXPECTED;
     fillin.string = expected[0]; // copy expected
     expected.pop_front();
     keys.pop_front();
-    if (keys.empty()) {
-      is_done = true; // so we get DONE next time
-    }
   } else {
+    // #1 First, we process each key
     fillin.type = LDML_ACTION_KEY_EVENT;
     fillin.k = next_key();
   }
 }
 
-
 key_event
 LdmlEmbeddedTestSource::next_key() {
   // mutate this->keys
-  return next_key(keys[0]);
+  return parse_next_key(keys[0]);
 }
 
 key_event
-LdmlEmbeddedTestSource::next_key(std::string &keys) {
+LdmlEmbeddedTestSource::parse_next_key(std::string &keys) {
   // Parse the next element of the string, chop it off, and return it
   // mutates keys
   if (keys.length() == 0)
@@ -448,10 +512,10 @@ LdmlEmbeddedTestSource::next_key(std::string &keys) {
 
 class LdmlJsonTestSource : public LdmlTestSource {
 public:
-  LdmlJsonTestSource(const std::string &path, km::core::kmx::kmx_plus *kmxplus);
+  LdmlJsonTestSource(const std::string &path);
   virtual ~LdmlJsonTestSource();
   virtual const std::u16string &get_context();
-  int load(const nlohmann::json &test);
+  int load(const nlohmann::json &test, const km::core::path &compiled);
   virtual void next_action(ldml_action &fillin);
 private:
   std::string path;
@@ -462,14 +526,14 @@ private:
    * Which action are we on?
   */
   std::size_t action_index = -1;
-  const km::core::kmx::kmx_plus *kmxplus;
   /** @return false if not found */
   bool set_key_from_id(key_event& k, const std::u16string& id);
   bool loaded_context = false;
+  bool check_keys = true;
 };
 
-LdmlJsonTestSource::LdmlJsonTestSource(const std::string &path, km::core::kmx::kmx_plus *k)
-:path(path), kmxplus(k) {
+LdmlJsonTestSource::LdmlJsonTestSource(const std::string &path)
+:path(path) {
 
 }
 
@@ -514,6 +578,13 @@ bool LdmlJsonTestSource::set_key_from_id(key_event& k, const std::u16string& id)
 void
 LdmlJsonTestSource::next_action(ldml_action &fillin) {
   if ((action_index+1) >= data["/actions"_json_pointer].size()) {
+    // add check keylist
+    if (check_keys) {
+      fillin.type = LDML_ACTION_CHECK_KEYLIST;
+      fillin.string.clear();
+      check_keys = false;
+      return;
+    }
     // at end, done
     fillin.type = LDML_ACTION_DONE;
     return;
@@ -581,19 +652,21 @@ LdmlJsonTestSource::get_context() {
   return context;
 }
 
-int LdmlJsonTestSource::load(const nlohmann::json &data) {
+int LdmlJsonTestSource::load(const nlohmann::json &data, const km::core::path &compiled) {
   this->data        = data;
-  // TODO-LDML: validate here?
-  return 0;
+  // TODO-LDML: validate JSON here?
+
+  // load up the kmx_plus
+  return load_kmx_plus(compiled);
 }
 
 #if defined(HAVE_ICU4C)
 class LdmlJsonRepertoireTestSource : public LdmlTestSource {
 public:
-  LdmlJsonRepertoireTestSource(const std::string &path, km::core::kmx::kmx_plus *kmxplus);
+  LdmlJsonRepertoireTestSource(const std::string &path );
   virtual ~LdmlJsonRepertoireTestSource();
   virtual const std::u16string &get_context();
-  int load(const nlohmann::json &test);
+  int load(const nlohmann::json &test, const km::core::path &compiled);
   virtual void next_action(ldml_action &fillin);
 private:
   std::string path;
@@ -605,11 +678,10 @@ private:
   std::unique_ptr<icu::UnicodeSet> uset;
   std::unique_ptr<icu::UnicodeSetIterator> iterator;
   bool need_check = false; // set this after each char
-  const km::core::kmx::kmx_plus *kmxplus;
 };
 
-LdmlJsonRepertoireTestSource::LdmlJsonRepertoireTestSource(const std::string &path, km::core::kmx::kmx_plus *k)
-:path(path), kmxplus(k){
+LdmlJsonRepertoireTestSource::LdmlJsonRepertoireTestSource(const std::string &path)
+:path(path) {
 
 }
 
@@ -697,7 +769,7 @@ LdmlJsonRepertoireTestSource::get_context() {
   return context; // no context needed
 }
 
-int LdmlJsonRepertoireTestSource::load(const nlohmann::json &data) {
+int LdmlJsonRepertoireTestSource::load(const nlohmann::json &data, const km::core::path &compiled) {
   this->data = data;  // TODO-LDML
   // Need an update to json.hpp to use contains()
   // if (data.contains("/type"_json_pointer)) {
@@ -728,7 +800,7 @@ int LdmlJsonRepertoireTestSource::load(const nlohmann::json &data) {
   // }
   // #endif
   iterator = std::unique_ptr<icu::UnicodeSetIterator>(new icu::UnicodeSetIterator(*uset));
-  return 0;
+  return load_kmx_plus(compiled);
 }
 #endif // HAVE_ICU4C
 
@@ -752,25 +824,6 @@ int LdmlJsonTestSourceFactory::load(const km::core::path &compiled, const km::co
     return __LINE__; // empty
   }
 
-  // check and load the KMX (yes, once again)
-  if(!km::core::ldml_processor::is_kmxplus_file(compiled, rawdata)) {
-    std::cerr << "Reading KMX for test purposes failed: " << compiled << std::endl;
-    return __LINE__;
-  }
-
-  auto comp_keyboard = (const km::core::kmx::COMP_KEYBOARD*)rawdata.data();
-  // initialize the kmxplus object with our copy
-  kmxplus.reset(new km::core::kmx::kmx_plus(comp_keyboard, rawdata.size()));
-
-  if (!kmxplus->is_valid()) {
-    std::cerr << "kmx_plus invalid" << std::endl;
-    return __LINE__;
-  }
-
-  if (!kmxplus->key2Helper.valid()) {
-    std::cerr << "kmx_plus invalid" << std::endl;
-    return __LINE__;
-  }
 
   auto conformsTo = data["/keyboardTest3/conformsTo"_json_pointer].get<std::string>();
   assert_or_return(std::string(LDML_CLDR_TEST_VERSION_LATEST) == conformsTo);
@@ -793,8 +846,8 @@ int LdmlJsonTestSourceFactory::load(const km::core::path &compiled, const km::co
       test_path.append(info_name).append("/tests/").append(tests_name).append("/").append(test_name);
       // std::cout << "JSON: reading " << info_name << "/" << test_path << std::endl;
 
-      std::unique_ptr<LdmlJsonTestSource> subtest(new LdmlJsonTestSource(test_path, kmxplus.get()));
-      assert_or_return(subtest->load(test) == 0);
+      std::unique_ptr<LdmlJsonTestSource> subtest(new LdmlJsonTestSource(test_path));
+      assert_or_return(subtest->load(test, compiled) == 0);
       test_map[test_path] = std::unique_ptr<LdmlTestSource>(subtest.release());
     }
   }
@@ -809,8 +862,8 @@ int LdmlJsonTestSourceFactory::load(const km::core::path &compiled, const km::co
     std::string test_path;
     test_path.append(info_name).append("/repertoire/").append(rep_name);
 
-    std::unique_ptr<LdmlJsonRepertoireTestSource> reptest(new LdmlJsonRepertoireTestSource(test_path, kmxplus.get()));
-    assert_or_return(reptest->load(rep) == 0);
+    std::unique_ptr<LdmlJsonRepertoireTestSource> reptest(new LdmlJsonRepertoireTestSource(test_path));
+    assert_or_return(reptest->load(rep, compiled) == 0);
     test_map[test_path] = std::unique_ptr<LdmlTestSource>(reptest.release());
   }
 
